#include "tp1_ex2_complex.hpp"
#include <iostream>
#include <cmath>

complex::complex(double r,double i) : re(r), im(i)
{ }

complex::complex(double r) : re(r), im(0.0)
{ }

complex::complex(const complex &c)
{
    copy(c);
}

void complex::copy(const complex &c)
{
    if (this == &c)
        return;
    this->re = c.re;
    this->im = c.im;
}

void complex::operator =(const complex& c)
{
    copy(c);
}

complex* complex::clone() const
{
    return new complex(*this);
}

double complex::real() const
{
    return re;
}

double complex::imag() const
{
    return im;
}

complex& complex::operator +=(complex a)
{
    re += a.re;
    im += a.im;
    return *this;
}

complex& complex::operator -=(complex a)
{
    re -= a.re;
    im -= a.im;
    return *this;
}

complex& complex::operator *=(complex a)
{
    re = re*a.re - im*a.im;
    im = re*a.im + im*a.re;
    return *this;
}

complex& complex::operator /=(complex a)
{
    double div = a.re*a.re + a.im*a.im;
    re = re*a.re + im*a.im;
    re /= div;
    im = -re*a.im + im*a.re;
    im /= div;
    return *this;
}

complex operator +(complex a, complex b)
{
    complex r = a;
    return r += b;
}

complex operator -(complex a, complex b)
{
    complex r = a;
    return r -= b;
}

complex operator *(complex a, complex b)
{
    complex r = a;
    return r *= b;
}

complex operator /(complex a, complex b)
{
    complex r = a;
    return r /= b;
}

complex operator -(complex a)
{
    complex r;
    return r -= a;
}

complex operator +(complex a)
{
    complex r;
    return r += a;
}

bool operator ==(complex a,complex b)
{
    return a.real() == b.real() && a.imag() == b.imag();
}

bool operator !=(complex a,complex b)
{
    return a.real() != b.real() || a.imag() != b.imag();
}

complex polar(double rho, double theta)
{
    complex r(rho*std::cos(theta), rho*std::sin(theta));
    return r;
}

complex conj(complex a)
{
    complex r(a.real(), -a.imag());
    return r;
}

double abs(complex a)
{
    double n = a.real()*a.real() + a.imag()*a.imag();
    n = std::sqrt(n);
    return n;
}

double arg(complex a)
{
    return std::atan2(a.imag(), a.real());
}

double norm(complex a)
{
    return abs(a);
}

double real(complex a)
{
    return a.real();
}

double imag(complex a)
{
    return a.imag();
}

std::ostream& operator<<(std::ostream& out, const complex& a)
{
    return out << '(' << a.real() << ',' << a.imag() << ')';
}

std::istream& operator >>(std::istream& in,complex& a)
{
    double re = 0.0, im = 0.0;
    char ch = ' ';
    in >> ch;
    if ( ch == '(' ) {
        in >> re >> ch;
        if ( ch == ',' ) {
            in >> im >> ch;
            if ( ch == ')' ) {
                a = complex(re, im);
            } else
                in.setstate(std::ios_base::failbit);
        } else if ( ch == ')' ) {
            a = re;
        } else
            in.setstate(std::ios_base::failbit);
    } else {
        in.putback(ch);
        in >> re;
        a = re;
    }
    return in;
}
