#ifndef TP2_EX1_LIST
#define TP2_EX1_LIST
template<class T>
class List
{
//Attributes
public:
    int size() const;
    // returns the list size

// Operations
public:
    virtual void add(T* item);
    // add the specified item at the end of list

    void add(int at_position, T* item);
    // add the specified item at specified position

    void remove(int at_position);
    // remove and deletes the item stored at specified position

    T* get_item(int at_position) const;
    // returns the item stored at specified position

// Implementation
public:
    List();
    virtual ~List();

    class invalid_position {};
    class out_of_memory {};
    // supported exceptions

protected:
    class T_Node
    {
    public:
        T_Node(T* item=NULL, T_Node* next_node=NULL, T_Node* prev_node=NULL)
        {
            _item = item;
            _next_node = next_node;
            _prev_node = prev_node;
        }

        ~T_Node()
        {
            delete _item;
        }

        T* _item;
        T_Node* _next_node;
        T_Node* _prev_node;
    };
    // this class is used internally to implement a single linked list

    T_Node* _first_node;
    T_Node* _last_node;
    // first node of the linked list

    int _number_of_nodes;
    // number of items stored in list

    T_Node* find_node(int node_position) const;
    // helper function to retrieve an indexed node

};

template<class T>
class SortedList : public List<T>
{
public:
    SortedList();
    ~SortedList();
    void add(T* item);
};

/////////////////////////////////////////////////////////////////////////////

template<class T>
List<T>::List()
{
    // initializes the list object for the first time

    _last_node = _first_node = NULL;
    _number_of_nodes = 0;
}

template<class T>
List<T>::~List()
{
    // deletes all stored items
    while (_number_of_nodes > 0)
        remove(0);
}

template<class T>
inline int List<T>::size() const
{
    // returns list size
    return _number_of_nodes;
}

template<class T>
inline void List<T>::add(T* item)
{
    // adds an item at list tail
    add(_number_of_nodes, item);
}

template<class T>
void List<T>::add(int at_position, T* item)
{
    // adds an item at specified position
    // throws an out_of_memory exception in case of low system resources
    //  or an invalid_position if insertion at specified position is
    //  impossible
    if (at_position == 0)
    {
        T_Node* nw_node = new T_Node(item, _first_node, NULL);

        if (nw_node == NULL)
            throw out_of_memory();
        else
        {
            if ( _first_node != NULL )
                _first_node->_prev_node  = nw_node;
            else
                _last_node = nw_node;
            _first_node = nw_node;
            _number_of_nodes++;
        }
    }
    else if (at_position == _number_of_nodes)
    {
        T_Node* nw_node = new T_Node(item, NULL, _last_node);

        if (nw_node == NULL)
            throw out_of_memory();
        else
        {
            _last_node->_next_node = nw_node;
            _last_node = nw_node;
            _number_of_nodes++;
        }

    }
    else
    {
        T_Node* previous_node = find_node(at_position - 1);
        // NOTE: find_node() can throw an invalid_position exception

        T_Node* next_node = previous_node->_next_node;
        T_Node* nw_node = new T_Node(item, next_node, previous_node);

        if (nw_node == NULL)
            throw out_of_memory();
        else
        {
            previous_node->_next_node =  nw_node;
            next_node->_prev_node = nw_node;
            _number_of_nodes++;
        }
    }
}

template<class T>
void List<T>::remove(int at_position)
{
    // removes and deletes the item indexed by the at_position value
    // throws an invalid_position if no item is available at specified
    //  position

    T_Node* node_to_delete;
    if (at_position == 0)
    {
        node_to_delete = _first_node;
        _first_node = node_to_delete->_next_node;
        if ( _first_node != NULL )
        _first_node->_prev_node = NULL;
    }
    else if (at_position == _number_of_nodes - 1)
    {
        node_to_delete = _last_node;
        _last_node = node_to_delete->_prev_node;
        _last_node->_next_node = NULL;
    }
    else
    {
        T_Node* previous_node = find_node(at_position - 1);
        // NOTE: find_node() can throw an invalid_position exception

        node_to_delete = previous_node->_next_node;
        previous_node->_next_node = node_to_delete->_next_node;
        node_to_delete->_next_node->_prev_node = previous_node;
    }

    delete node_to_delete;
    _number_of_nodes--;
}

template<class T>
inline T* List<T>::get_item(int at_position) const
{
    // returns the requested item
    // throws an invalid_position if no item is available at specified
    //  position

    T_Node* node = find_node(at_position);
    return node->_item;
}

template<class T>
typename List<T>::T_Node* List<T>::find_node(int node_position) const
{
    // searches (by index) a node in the internal single linked list
    // throws an invalid_position exception if no item found

    if (node_position < 0)
        throw invalid_position();

    T_Node* node = _first_node;
    int node_index = 0;

    while (node_index < node_position)
    {
        if (node == NULL)
            throw invalid_position();

        node = node->_next_node;
        node_index++;
    }

    return node;
}

template<class T>
T max (const List<T> &list)
{
    T max = *list.get_item(0);
    for ( int i = 1 ; i < list.size() ; ++i )
    {
        if ( *list.get_item(i) > max )
            max = *list.get_item(i);
    }
    return max;
}

/////////////////////////////////////////////////////////////////////////////

template<class T>
SortedList<T>::SortedList() : List<T>()
{
}

template<class T>
SortedList<T>::~SortedList()
{
}

template<class T>
void SortedList<T>::add(T* item)
{
    typename List<T>::T_Node* node = this->_first_node;
    int cnt = 0;

    while ((node != NULL) && (*node->_item < *item))
    {
        node = node->_next_node;
        ++cnt;
    }
    List<T>::add(cnt, item);
}

#endif // TP2_EX1_LIST
